#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <ncurses.h>
#include <unistd.h>
#include <termios.h>
#include <limits.h>
#include <time.h>
#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>
#include <glad/glad.h>
#include <GLFW/glfw3.h>

#include "hash.h"
#include "structure.h"
#include "base.h"
#include "move.h"
#include "triangles.h"
#include "proj.h"
#include "entities.h"
#include "display.h"
#include "generation.h"

double sim_time ;

void processInput(GLFWwindow *window, float dtime) {
    if (glfwGetKey(window, GLFW_KEY_ESCAPE) == GLFW_PRESS) {
        glfwSetWindowShouldClose(window, true);
    }
    if(glfwGetKey(window, GLFW_KEY_S) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camz -= speed*cos(rot_hz)*cos(rot_vt)/10;
            camx -= speed*sin(rot_hz)*cos(rot_vt)/10;
            camy += speed*sin(rot_vt)/10;
            if(is_colliding(dtime)) {
                camz += speed*cos(rot_hz)*cos(rot_vt)/10;
                camx += speed*sin(rot_hz)*cos(rot_vt)/10;
                camy -= speed*sin(rot_vt)/10;
                k=11;
            }
        }
    }
    if(glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camx -= speed*cos(rot_hz)/10;
            camz += speed*sin(rot_hz)/10;
            if(is_colliding(dtime)) {
                camx += speed*cos(rot_hz)/10;
                camz -= speed*sin(rot_hz)/10;
                k=11;
            }
        }
    }
    if(glfwGetKey(window, GLFW_KEY_Z) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camz += speed*cos(rot_hz)*cos(rot_vt)/10;
            camx += speed*sin(rot_hz)*cos(rot_vt)/10;
            camy -= speed*sin(rot_vt)/10;
            if(is_colliding(dtime)) {
                camz -= speed*cos(rot_hz)*cos(rot_vt)/10;
                camx -= speed*sin(rot_hz)*cos(rot_vt)/10;
                camy += speed*sin(rot_vt)/10;
                k=11;
            }
        }
    }
    if(glfwGetKey(window, GLFW_KEY_Q) == GLFW_PRESS) {
        for(int k = 0; k < 10; k++) {
            camx += speed*cos(rot_hz)/10;
            camz -= speed*sin(rot_hz)/10;
            if(is_colliding(dtime)) {
                camx -= speed*cos(rot_hz)/10;
                camz += speed*sin(rot_hz)/10;
                k=11;
            }
        }
    }

    if(glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS) {
        stop_evetything = true ;
    }

    if(glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS) {
        rot_hz -= sensitivity;
    }
    if(glfwGetKey(window, GLFW_KEY_E) == GLFW_PRESS) {
        rot_hz += sensitivity;
    }

    if(glfwGetKey(window, GLFW_KEY_P) == GLFW_PRESS) {
        rot_vt -= sensitivity;
    }
    if(glfwGetKey(window, GLFW_KEY_M) == GLFW_PRESS) {
        rot_vt += sensitivity;
    }
}

const char *vertexShaderSource = "#version 330 core\n"
    "layout (location = 0) in vec3 aPos;\n"
    "uniform mat4 model;\n"
    "uniform mat4 view;\n"
    "uniform mat4 projection;\n"
    "void main() {\n"
    "   gl_Position = projection * view * model * vec4(aPos, 1.0);\n"
    "}\0";

// Fragment Shader Source
const char *fragmentShaderSource = "#version 330 core\n"
    "out vec4 FragColor;\n"
    "void main() {\n"
    "   FragColor = vec4(1.0, 0.5, 0.2, 1.0);\n"
    "}\0";

int main_alt() {
    // glfw: initialize and configure
    // ------------------------------
    glfwInit();
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

    // glfw window creation
    // --------------------
    GLFWwindow* window = glfwCreateWindow(1500, 1000, "LearnOpenGL", NULL, NULL);
    if (window == NULL)
    {
        fprintf(stderr, "ERROR : cannot initialize GL window");
        glfwTerminate();
        return -1;
    }
    glfwMakeContextCurrent(window);

    // glad: load all OpenGL function pointers
    // ---------------------------------------
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        fprintf(stderr, "ERROR : cannot initialize GLAD loader");
        return -1;
    }

    //printf("%f\n", glDepthRange);

    // build and compile our shader program
    // ------------------------------------
    // vertex shader
    unsigned int vertexShader = glCreateShader(GL_VERTEX_SHADER);
    glShaderSource(vertexShader, 1, &vertexShaderSource, NULL);
    glCompileShader(vertexShader);
    // check for shader compile errors
    int success;
    char infoLog[512];
    glGetShaderiv(vertexShader, GL_COMPILE_STATUS, &success);
    if (!success)
    {
        glGetShaderInfoLog(vertexShader, 512, NULL, infoLog);
        fprintf(stderr, "ERROR : cannot initialize shader (1)");
    }
    // fragment shader
    unsigned int fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);
    glShaderSource(fragmentShader, 1, &fragmentShaderSource, NULL);
    glCompileShader(fragmentShader);
    // check for shader compile errors
    glGetShaderiv(fragmentShader, GL_COMPILE_STATUS, &success);
    if (!success)
    {
        glGetShaderInfoLog(fragmentShader, 512, NULL, infoLog);
        fprintf(stderr, "ERROR : cannot initialize shader (2)");
    }
    // link shaders
    unsigned int shaderProgram = glCreateProgram();
    glAttachShader(shaderProgram, vertexShader);
    glAttachShader(shaderProgram, fragmentShader);
    glLinkProgram(shaderProgram);
    // check for linking errors
    glGetProgramiv(shaderProgram, GL_LINK_STATUS, &success);
    if (!success) {
        glGetProgramInfoLog(shaderProgram, 512, NULL, infoLog);
        fprintf(stderr, "ERROR : cannot link shaders");
    }
    glDeleteShader(vertexShader);
    glDeleteShader(fragmentShader);

    // set up data (buffer(s)) and configure vertex attributes
    // ------------------------------------------------------------------

    unsigned int VBO, VAO;
    glGenVertexArrays(1, &VAO);
    glGenBuffers(1, &VBO);
    // bind the Vertex Array Object first, then bind and set vertex buffer(s), and then configure vertex attributes(s).
    glBindVertexArray(VAO);


    // uncomment this call to draw in wireframe polygons.
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

    // render loop
    // -----------
    init_csts();
    init_hashtbl();
    init_draworder();
    init_ent_generator(10);
    trInit();
    init_proj();
    parse_rooms(5);

    clock_t origin = clock();
    clock_t finish = clock();
    while (!glfwWindowShouldClose(window)) {
        // input
        // -----
        glClearColor(0.2f, 0.3f, 0.3f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT);
        origin = clock();

        generate_nearby_chunks(1);

        //gl_renderTriangle(shaderProgram, VAO, VBO, -0.5, -0.5, 0.0, 0.5, -0.5, 0.0, 0.0, 0.5, 0.0, 1.0, 1.0, 1.0);
        gl_renderAll(shaderProgram, VAO, VBO);

        //printf("01\n");
        //fflush(stdout);

        // glfw: swap buffers and poll IO events (keys pressed/released, mouse moved etc.)
        // -------------------------------------------------------------------------------
        finish = clock();
        processInput(window, ((float)origin - (float)finish)/CLOCKS_PER_SEC);
        update_entities(((float)origin - (float)finish)/CLOCKS_PER_SEC);
        updateAllProj(((float)origin - (float)finish)/CLOCKS_PER_SEC);
        glfwSwapBuffers(window);
        glfwPollEvents();
    }

    for(int k = 0; k < MAX_SIZE; k++) {
        free(triangles_to_render[k]);
        free(triangles_og_coords[k]);
    }
    free(drawOrder);
    free(triangles_to_render);
    free(triangles_og_coords);
    free(reds);
    free(greens);
    free(blues);
    free(triangles_order);
    free(triangles_shr);
    free(visited_tri);
    free_proj();
    //printf("10\n");
    //fflush(stdout);
    hashtbl_free(visited);
    free_pool();

    // optional: de-allocate all resources once they've outlived their purpose:
    // ------------------------------------------------------------------------
    glDeleteVertexArrays(1, &VAO);
    glDeleteBuffers(1, &VBO);
    glDeleteProgram(shaderProgram);

    // glfw: terminate, clearing all previously allocated GLFW resources.
    // ------------------------------------------------------------------
    glfwTerminate();
    return 0;
}

int main(int argc, char** argv) {
    srand(time(NULL));
    return main_alt();

    //-------------------------------------------------------------------------------//
    
    if (SDL_Init(SDL_INIT_EVERYTHING) != 0) {
        printf( "error initializing SDL: %s\n", SDL_GetError());
    }
    SDL_Window* win = SDL_CreateWindow("Game",
                                       SDL_WINDOWPOS_CENTERED,
                                       SDL_WINDOWPOS_CENTERED,
                                       1500, 1000, 0);
 
    Uint32 render_flags = SDL_RENDERER_ACCELERATED;
    SDL_Renderer* rend = SDL_CreateRenderer(win, -1, render_flags);
    if(rend == NULL) {
        printf( "ERROR : cannot initialize SDL renderer\n");
        exit(1);
    }
    printf( "%d\n", SDL_SetRenderDrawBlendMode(rend, SDL_BLENDMODE_BLEND));
    
    //-------------------------------------------------------------------------------//

    if(SDL_Init(SDL_INIT_AUDIO)) {
        fprintf(stderr, "cannot initialize audio");
        exit(1);
    }
    SDL_SetRelativeMouseMode(true) ;

    /* -------------------------------------------------------- */
    int fps = 60;
    int interval = 1000000/fps;
    double intervalf = 1.0/((double)(fps));

    bool debug_main = true ;

    init_csts();
    init_hashtbl();
    init_draworder();
    init_ent_generator(10);
    trInit();
    init_proj();
    parse_rooms(5);
    import_digits(rend) ;
    import_letters(rend) ;
    sim_time = 0.0 ;
    clock_t origin = clock();
    clock_t finish = clock();

    clock_t entstart = clock();
    clock_t entend = clock();
    float delta;
    while(!stop_evetything) {
        resetRenderer(rend) ;

        origin = clock();
        SDL_SetRenderDrawColor(rend, 255, 255, 255, SDL_ALPHA_OPAQUE) ;
        entend = clock();
        //printf("00\n");
        //printf("%s\n", SDL_GetError());
        //fflush(stdout);
        playerActions(((float)entend - (float)entstart)/CLOCKS_PER_SEC) ;
        //printf("01\n");
        //fflush(stdout);
        generate_nearby_chunks(1);
        //printf("02\n");
        //fflush(stdout);

        entend = clock();
        update_entities(((float)entend - (float)entstart)/CLOCKS_PER_SEC);
        updateAllProj(((float)entend - (float)entstart)/CLOCKS_PER_SEC);
        //printf("03\n");
        //fflush(stdout);
        entstart = clock();

        //printf("-->%d\n", triangles_i);
        drawCurrentRoom(rend);
        //printf("-->%d\n", triangles_i);
        //printf("-->%d\n", triangles_i);
        //printf("04\n");
        //printf("%s\n", SDL_GetError());
        //fflush(stdout);
        drawData(rend) ;
        //printf("05\n");
        //fflush(stdout);
        drawHPbar(rend);
        //printf("06\n");
        //fflush(stdout);
        finish = clock();
        fade_dmg = max(fade_dmg-5, 0);
        delta = ((float)finish - (float)origin)/CLOCKS_PER_SEC;
        //printf("07\n");
        //fflush(stdout);
        drawNumberToRenderer(rend, digits, (int)(1.0f/delta), 720, 60, 75/2, 105/2, 0);
        drawNumberToRenderer(rend, digits, (int)(10*sim_time), 720, 110, 75/2, 105/2, 0);
        drawNumberToRenderer(rend, digits, coins, 1500/2-55, 1000 - 70, 75/3, 105/3, 0);
        //printf("08\n");
        //fflush(stdout);
        //printf("%s\n", SDL_GetError());
        updateRenderer(rend) ;
        sim_time += delta + intervalf ;
        //printf("09\n");
        //fflush(stdout);
        if(player_hp <= 0) {
            stop_evetything = true ;
        }
        usleep(interval) ;
    }
    //printf("GPNE\n");
    //fflush(stdout);
    free_digits(digits) ;

    for(int k = 0; k < MAX_SIZE; k++) {
        free(triangles_to_render[k]);
        free(triangles_og_coords[k]);
    }
    free(drawOrder);
    free(triangles_to_render);
    free(triangles_og_coords);
    free(reds);
    free(greens);
    free(blues);
    free(triangles_order);
    free(triangles_shr);
    free(visited_tri);
    free_proj();
    //printf("10\n");
    //fflush(stdout);
    hashtbl_free(visited);
    free_pool();

    /* -------------------------------------------------------- */

    SDL_DestroyRenderer(rend);
    //printf("10\n");
    //fflush(stdout);
    SDL_DestroyWindow(win);
    //printf("11\n");
    //fflush(stdout);
    SDL_Quit();
    //printf("12\n");
    //fflush(stdout);

    /* -------------------------------------------------------- */

    //printf("Done\n") ;
    //fflush(stdout);
    return 0;
}